@@ -1,7 +1,7 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; frozen_string_literal: true -*-
 #
 #--
-# Copyright (C) 2009-2016 Thomas Leitner <t_leitner@gmx.at>
+# Copyright (C) 2009-2019 Thomas Leitner <t_leitner@gmx.at>
 #
 # This file is part of kramdown which is licensed under the MIT.
 #++
@@ -46,15 +46,16 @@ module Kramdown
         @toc_code = nil
         @indent = 2
         @stack = []
-      end
 
-      # The mapping of element type to conversion method.
-      DISPATCHER = Hash.new {|h,k| h[k] = "convert_#{k}"}
+        # stash string representation of symbol to avoid allocations from multiple interpolations.
+        @highlighter_class = " highlighter-#{options[:syntax_highlighter]}"
+        @dispatcher = Hash.new {|h, k| h[k] = :"convert_#{k}" }
+      end
 
       # Dispatch the conversion of the element +el+ to a +convert_TYPE+ method using the +type+ of
       # the element.
       def convert(el, indent = -@indent)
-        send(DISPATCHER[el.type], el, indent)
+        send(@dispatcher[el.type], el, indent)
       end
 
       # Return the converted content of the children of +el+ as a string. The parameter +indent+ has
@@ -63,32 +64,48 @@ module Kramdown
       # Pushes +el+ onto the @stack before converting the child elements and pops it from the stack
       # afterwards.
       def inner(el, indent)
-        result = ''
+        result = +''
         indent += @indent
         @stack.push(el)
         el.children.each do |inner_el|
-          result << send(DISPATCHER[inner_el.type], inner_el, indent)
+          result << send(@dispatcher[inner_el.type], inner_el, indent)
         end
         @stack.pop
         result
       end
 
-      def convert_blank(el, indent)
+      def convert_blank(_el, _indent)
         "\n"
       end
 
-      def convert_text(el, indent)
-        escape_html(el.value, :text)
+      def convert_text(el, _indent)
+        escaped = escape_html(el.value, :text)
+        @options[:remove_line_breaks_for_cjk] ? fix_cjk_line_break(escaped) : escaped
       end
 
       def convert_p(el, indent)
         if el.options[:transparent]
           inner(el, indent)
+        elsif el.children.size == 1 && el.children.first.type == :img &&
+            el.children.first.options[:ial]&.[](:refs)&.include?('standalone')
+          convert_standalone_image(el.children.first, indent)
         else
-          format_as_block_html(el.type, el.attr, inner(el, indent), indent)
+          format_as_block_html("p", el.attr, inner(el, indent), indent)
         end
       end
 
+      # Helper method used by +convert_p+ to convert a paragraph that only contains a single :img
+      # element.
+      def convert_standalone_image(el, indent)
+        attr = el.attr.dup
+        figure_attr = {}
+        figure_attr['class'] = attr.delete('class') if attr.key?('class')
+        figure_attr['id'] = attr.delete('id') if attr.key?('id')
+        body = "#{' ' * (indent + @indent)}<img#{html_attributes(attr)} />\n" \
+          "#{' ' * (indent + @indent)}<figcaption>#{attr['alt']}</figcaption>\n"
+        format_as_indented_block_html("figure", figure_attr, body, indent)
+      end
+
       def convert_codeblock(el, indent)
         attr = el.attr.dup
         lang = extract_code_language!(attr)
@@ -97,7 +114,7 @@ module Kramdown
 
         if highlighted_code
           add_syntax_highlighter_to_class_attr(attr, lang || hl_opts[:default_lang])
-          "#{' '*indent}<div#{html_attributes(attr)}>#{highlighted_code}#{' '*indent}</div>\n"
+          "#{' ' * indent}<div#{html_attributes(attr)}>#{highlighted_code}#{' ' * indent}</div>\n"
         else
           result = escape_html(el.value)
           result.chomp!
@@ -114,12 +131,13 @@ module Kramdown
           end
           code_attr = {}
           code_attr['class'] = "language-#{lang}" if lang
-          "#{' '*indent}<pre#{html_attributes(attr)}><code#{html_attributes(code_attr)}>#{result}\n</code></pre>\n"
+          "#{' ' * indent}<pre#{html_attributes(attr)}>" \
+            "<code#{html_attributes(code_attr)}>#{result}\n</code></pre>\n"
         end
       end
 
       def convert_blockquote(el, indent)
-        format_as_indented_block_html(el.type, el.attr, inner(el, indent), indent)
+        format_as_indented_block_html("blockquote", el.attr, inner(el, indent), indent)
       end
 
       def convert_header(el, indent)
@@ -133,62 +151,68 @@ module Kramdown
       end
 
       def convert_hr(el, indent)
-        "#{' '*indent}<hr#{html_attributes(el.attr)} />\n"
+        "#{' ' * indent}<hr#{html_attributes(el.attr)} />\n"
       end
 
+      ZERO_TO_ONETWENTYEIGHT = (0..128).to_a.freeze
+      private_constant :ZERO_TO_ONETWENTYEIGHT
+
       def convert_ul(el, indent)
-        if !@toc_code && (el.options[:ial][:refs].include?('toc') rescue nil)
-          @toc_code = [el.type, el.attr, (0..128).to_a.map{|a| rand(36).to_s(36)}.join]
+        if !@toc_code && el.options.dig(:ial, :refs)&.include?('toc')
+          @toc_code = [el.type, el.attr, ZERO_TO_ONETWENTYEIGHT.map { rand(36).to_s(36) }.join]
           @toc_code.last
-        elsif !@footnote_location && el.options[:ial] && (el.options[:ial][:refs] || []).include?('footnotes')
-          @footnote_location = (0..128).to_a.map{|a| rand(36).to_s(36)}.join
+        elsif !@footnote_location && el.options.dig(:ial, :refs)&.include?('footnotes')
+          @footnote_location = ZERO_TO_ONETWENTYEIGHT.map { rand(36).to_s(36) }.join
         else
           format_as_indented_block_html(el.type, el.attr, inner(el, indent), indent)
         end
       end
-      alias :convert_ol :convert_ul
+      alias convert_ol convert_ul
 
       def convert_dl(el, indent)
-        format_as_indented_block_html(el.type, el.attr, inner(el, indent), indent)
+        format_as_indented_block_html("dl", el.attr, inner(el, indent), indent)
       end
 
       def convert_li(el, indent)
-        output = ' '*indent << "<#{el.type}" << html_attributes(el.attr) << ">"
+        output = ' ' * indent << "<#{el.type}" << html_attributes(el.attr) << ">"
         res = inner(el, indent)
         if el.children.empty? || (el.children.first.type == :p && el.children.first.options[:transparent])
-          output << res << (res =~ /\n\Z/ ? ' '*indent : '')
+          output << res << (res =~ /\n\Z/ ? ' ' * indent : '')
         else
-          output << "\n" << res << ' '*indent
+          output << "\n" << res << ' ' * indent
         end
         output << "</#{el.type}>\n"
       end
-      alias :convert_dd :convert_li
+      alias convert_dd convert_li
 
       def convert_dt(el, indent)
         attr = el.attr.dup
         @stack.last.options[:ial][:refs].each do |ref|
           if ref =~ /\Aauto_ids(?:-([\w-]+))?/
-            attr['id'] = ($1 ? $1 : '') << basic_generate_id(el.options[:raw_text])
+            attr['id'] = "#{$1}#{basic_generate_id(el.options[:raw_text])}".lstrip
             break
           end
         end if !attr['id'] && @stack.last.options[:ial] && @stack.last.options[:ial][:refs]
-        format_as_block_html(el.type, attr, inner(el, indent), indent)
+        format_as_block_html("dt", attr, inner(el, indent), indent)
       end
 
       def convert_html_element(el, indent)
         res = inner(el, indent)
         if el.options[:category] == :span
-          "<#{el.value}#{html_attributes(el.attr)}" << (res.empty? && HTML_ELEMENTS_WITHOUT_BODY.include?(el.value) ? " />" : ">#{res}</#{el.value}>")
+          "<#{el.value}#{html_attributes(el.attr)}" + \
+            (res.empty? && HTML_ELEMENTS_WITHOUT_BODY.include?(el.value) ? " />" : ">#{res}</#{el.value}>")
         else
-          output = ''
-          output << ' '*indent if @stack.last.type != :html_element || @stack.last.options[:content_model] != :raw
+          output = +''
+          if @stack.last.type != :html_element || @stack.last.options[:content_model] != :raw
+            output << ' ' * indent
+          end
           output << "<#{el.value}#{html_attributes(el.attr)}"
           if el.options[:is_closed] && el.options[:content_model] == :raw
             output << " />"
           elsif !res.empty? && el.options[:content_model] != :block
             output << ">#{res}</#{el.value}>"
           elsif !res.empty?
-            output << ">\n#{res.chomp}\n"  << ' '*indent << "</#{el.value}>"
+            output << ">\n#{res.chomp}\n" << ' ' * indent << "</#{el.value}>"
           elsif HTML_ELEMENTS_WITHOUT_BODY.include?(el.value)
             output << " />"
           else
@@ -200,21 +224,22 @@ module Kramdown
       end
 
       def convert_xml_comment(el, indent)
-        if el.options[:category] == :block && (@stack.last.type != :html_element || @stack.last.options[:content_model] != :raw)
-          ' '*indent << el.value << "\n"
+        if el.options[:category] == :block &&
+            (@stack.last.type != :html_element || @stack.last.options[:content_model] != :raw)
+          ' ' * indent << el.value << "\n"
         else
           el.value
         end
       end
-      alias :convert_xml_pi :convert_xml_comment
+      alias convert_xml_pi convert_xml_comment
 
       def convert_table(el, indent)
         format_as_indented_block_html(el.type, el.attr, inner(el, indent), indent)
       end
-      alias :convert_thead :convert_table
-      alias :convert_tbody :convert_table
-      alias :convert_tfoot :convert_table
-      alias :convert_tr  :convert_table
+      alias convert_thead convert_table
+      alias convert_tbody convert_table
+      alias convert_tfoot convert_table
+      alias convert_tr convert_table
 
       ENTITY_NBSP = ::Kramdown::Utils::Entities.entity('nbsp') # :nodoc:
 
@@ -225,32 +250,32 @@ module Kramdown
         alignment = @stack[-3].options[:alignment][@stack.last.children.index(el)]
         if alignment != :default
           attr = el.attr.dup
-          attr['style'] = (attr.has_key?('style') ? "#{attr['style']}; ": '') << "text-align: #{alignment}"
+          attr['style'] = (attr.key?('style') ? "#{attr['style']}; " : '') + "text-align: #{alignment}"
         end
         format_as_block_html(type, attr, res.empty? ? entity_to_str(ENTITY_NBSP) : res, indent)
       end
 
       def convert_comment(el, indent)
         if el.options[:category] == :block
-          "#{' '*indent}<!-- #{el.value} -->\n"
+          "#{' ' * indent}<!-- #{el.value} -->\n"
         else
           "<!-- #{el.value} -->"
         end
       end
 
-      def convert_br(el, indent)
+      def convert_br(_el, _indent)
         "<br />"
       end
 
       def convert_a(el, indent)
-        format_as_span_html(el.type, el.attr, inner(el, indent))
+        format_as_span_html("a", el.attr, inner(el, indent))
       end
 
-      def convert_img(el, indent)
+      def convert_img(el, _indent)
         "<img#{html_attributes(el.attr)} />"
       end
 
-      def convert_codespan(el, indent)
+      def convert_codespan(el, _indent)
         attr = el.attr.dup
         lang = extract_code_language(attr)
         hl_opts = {}
@@ -264,21 +289,24 @@ module Kramdown
         format_as_span_html('code', attr, result)
       end
 
-      def convert_footnote(el, indent)
+      def convert_footnote(el, _indent)
         repeat = ''
-        if (footnote = @footnotes_by_name[el.options[:name]])
+        name = @options[:footnote_prefix] + el.options[:name]
+        if (footnote = @footnotes_by_name[name])
           number = footnote[2]
           repeat = ":#{footnote[3] += 1}"
         else
           number = @footnote_counter
           @footnote_counter += 1
-          @footnotes << [el.options[:name], el.value, number, 0]
-          @footnotes_by_name[el.options[:name]] = @footnotes.last
+          @footnotes << [name, el.value, number, 0]
+          @footnotes_by_name[name] = @footnotes.last
         end
-        "<sup id=\"fnref:#{el.options[:name]}#{repeat}\"><a href=\"#fn:#{el.options[:name]}\" class=\"footnote\">#{number}</a></sup>"
+        "<sup id=\"fnref:#{name}#{repeat}\" role=\"doc-noteref\">" \
+          "<a href=\"#fn:#{name}\" class=\"footnote\">" \
+          "#{number}</a></sup>"
       end
 
-      def convert_raw(el, indent)
+      def convert_raw(el, _indent)
         if !el.options[:type] || el.options[:type].empty? || el.options[:type].include?('html')
           el.value + (el.options[:category] == :block ? "\n" : '')
         else
@@ -289,35 +317,41 @@ module Kramdown
       def convert_em(el, indent)
         format_as_span_html(el.type, el.attr, inner(el, indent))
       end
-      alias :convert_strong :convert_em
+      alias convert_strong convert_em
 
-      def convert_entity(el, indent)
+      def convert_entity(el, _indent)
         entity_to_str(el.value, el.options[:original])
       end
 
       TYPOGRAPHIC_SYMS = {
-        :mdash => [::Kramdown::Utils::Entities.entity('mdash')],
-        :ndash => [::Kramdown::Utils::Entities.entity('ndash')],
-        :hellip => [::Kramdown::Utils::Entities.entity('hellip')],
-        :laquo_space => [::Kramdown::Utils::Entities.entity('laquo'), ::Kramdown::Utils::Entities.entity('nbsp')],
-        :raquo_space => [::Kramdown::Utils::Entities.entity('nbsp'), ::Kramdown::Utils::Entities.entity('raquo')],
-        :laquo => [::Kramdown::Utils::Entities.entity('laquo')],
-        :raquo => [::Kramdown::Utils::Entities.entity('raquo')]
+        mdash: [::Kramdown::Utils::Entities.entity('mdash')],
+        ndash: [::Kramdown::Utils::Entities.entity('ndash')],
+        hellip: [::Kramdown::Utils::Entities.entity('hellip')],
+        laquo_space: [::Kramdown::Utils::Entities.entity('laquo'),
+                      ::Kramdown::Utils::Entities.entity('nbsp')],
+        raquo_space: [::Kramdown::Utils::Entities.entity('nbsp'),
+                      ::Kramdown::Utils::Entities.entity('raquo')],
+        laquo: [::Kramdown::Utils::Entities.entity('laquo')],
+        raquo: [::Kramdown::Utils::Entities.entity('raquo')],
       } # :nodoc:
-      def convert_typographic_sym(el, indent)
-        TYPOGRAPHIC_SYMS[el.value].map {|e| entity_to_str(e)}.join('')
+      def convert_typographic_sym(el, _indent)
+        if (result = @options[:typographic_symbols][el.value])
+          escape_html(result, :text)
+        else
+          TYPOGRAPHIC_SYMS[el.value].map {|e| entity_to_str(e) }.join('')
+        end
       end
 
-      def convert_smart_quote(el, indent)
+      def convert_smart_quote(el, _indent)
         entity_to_str(smart_quote_entity(el))
       end
 
       def convert_math(el, indent)
-        if (result = format_math(el, :indent => indent))
+        if (result = format_math(el, indent: indent))
           result
         else
           attr = el.attr.dup
-          (attr['class'] = (attr['class'] || '') << " kdmath").lstrip!
+          attr['class'] = "#{attr['class']} kdmath".lstrip
           if el.options[:category] == :block
             format_as_block_html('div', attr, "$$\n#{el.value}\n$$", indent)
           else
@@ -326,7 +360,7 @@ module Kramdown
         end
       end
 
-      def convert_abbreviation(el, indent)
+      def convert_abbreviation(el, _indent)
         title = @root.options[:abbrev_defs][el.value]
         attr = @root.options[:abbrev_attr][el.value].dup
         attr['title'] = title unless title.empty?
@@ -342,7 +376,7 @@ module Kramdown
         end
         if @toc_code
           toc_tree = generate_toc_tree(@toc, @toc_code[0], @toc_code[1] || {})
-          text = if toc_tree.children.size > 0
+          text = if !toc_tree.children.empty?
                    convert(toc_tree, 0)
                  else
                    ''
@@ -359,30 +393,30 @@ module Kramdown
 
       # Format the given element as block HTML.
       def format_as_block_html(name, attr, body, indent)
-        "#{' '*indent}<#{name}#{html_attributes(attr)}>#{body}</#{name}>\n"
+        "#{' ' * indent}<#{name}#{html_attributes(attr)}>#{body}</#{name}>\n"
       end
 
       # Format the given element as block HTML with a newline after the start tag and indentation
       # before the end tag.
       def format_as_indented_block_html(name, attr, body, indent)
-        "#{' '*indent}<#{name}#{html_attributes(attr)}>\n#{body}#{' '*indent}</#{name}>\n"
+        "#{' ' * indent}<#{name}#{html_attributes(attr)}>\n#{body}#{' ' * indent}</#{name}>\n"
       end
 
       # Add the syntax highlighter name to the 'class' attribute of the given attribute hash. And
       # overwrites or add a "language-LANG" part using the +lang+ parameter if +lang+ is not nil.
       def add_syntax_highlighter_to_class_attr(attr, lang = nil)
-        (attr['class'] = (attr['class'] || '') + " highlighter-#{@options[:syntax_highlighter]}").lstrip!
+        (attr['class'] = (attr['class'] || '') + @highlighter_class).lstrip!
         attr['class'].sub!(/\blanguage-\S+|(^)/) { "language-#{lang}#{$1 ? ' ' : ''}" } if lang
       end
 
       # Generate and return an element tree for the table of contents.
       def generate_toc_tree(toc, type, attr)
-        sections = Element.new(type, nil, attr)
+        sections = Element.new(type, nil, attr.dup)
         sections.attr['id'] ||= 'markdown-toc'
         stack = []
         toc.each do |level, id, children|
-          li = Element.new(:li, nil, nil, {:level => level})
-          li.children << Element.new(:p, nil, nil, {:transparent => true})
+          li = Element.new(:li, nil, nil, level: level)
+          li.children << Element.new(:p, nil, nil, transparent: true)
           a = Element.new(:a, nil)
           a.attr['href'] = "##{id}"
           a.attr['id'] = "#{sections.attr['id']}-#{id}"
@@ -391,7 +425,7 @@ module Kramdown
           li.children << Element.new(type)
 
           success = false
-          while !success
+          until success
             if stack.empty?
               sections.children << li
               stack << li
@@ -402,13 +436,13 @@ module Kramdown
               success = true
             else
               item = stack.pop
-              item.children.pop unless item.children.last.children.size > 0
+              item.children.pop if item.children.last.children.empty?
             end
           end
         end
-        while !stack.empty?
+        until stack.empty?
           item = stack.pop
-          item.children.pop unless item.children.last.children.size > 0
+          item.children.pop if item.children.last.children.empty?
         end
         sections
       end
@@ -438,17 +472,17 @@ module Kramdown
 
       # Obfuscate the +text+ by using HTML entities.
       def obfuscate(text)
-        result = ""
+        result = +''
         text.each_byte do |b|
-          result << (b > 128 ? b.chr : "&#%03d;" % b)
+          result << (b > 128 ? b.chr : sprintf("&#%03d;", b))
         end
         result.force_encoding(text.encoding)
         result
       end
 
-      FOOTNOTE_BACKLINK_FMT = "%s<a href=\"#fnref:%s\" class=\"reversefootnote\">%s</a>"
+      FOOTNOTE_BACKLINK_FMT = "%s<a href=\"#fnref:%s\" class=\"reversefootnote\" role=\"doc-backlink\">%s</a>"
 
-      # Return a HTML ordered list with the footnote content for the used footnotes.
+      # Return an HTML ordered list with the footnote content for the used footnotes.
       def footnote_content
         ol = Element.new(:ol)
         ol.attr['start'] = @footnote_start if @footnote_start != 1
@@ -456,29 +490,43 @@ module Kramdown
         backlink_text = escape_html(@options[:footnote_backlink], :text)
         while i < @footnotes.length
           name, data, _, repeat = *@footnotes[i]
-          li = Element.new(:li, nil, {'id' => "fn:#{name}"})
+          li = Element.new(:li, nil, 'id' => "fn:#{name}", 'role' => 'doc-endnote')
           li.children = Marshal.load(Marshal.dump(data.children))
 
-          if li.children.last.type == :p
-            para = li.children.last
+          para = nil
+          if li.children.last.type == :p || @options[:footnote_backlink_inline]
+            parent = li
+            while !parent.children.empty? && ![:p, :header].include?(parent.children.last.type)
+              parent = parent.children.last
+            end
+            para = parent.children.last
             insert_space = true
-          else
+          end
+
+          unless para
             li.children << (para = Element.new(:p))
             insert_space = false
           end
 
           unless @options[:footnote_backlink].empty?
             nbsp = entity_to_str(ENTITY_NBSP)
-            para.children << Element.new(:raw, FOOTNOTE_BACKLINK_FMT % [insert_space ? nbsp : '', name, backlink_text])
+            value = sprintf(FOOTNOTE_BACKLINK_FMT, (insert_space ? nbsp : ''), name, backlink_text)
+            para.children << Element.new(:raw, value)
             (1..repeat).each do |index|
-              para.children << Element.new(:raw, FOOTNOTE_BACKLINK_FMT % [nbsp, "#{name}:#{index}", "#{backlink_text}<sup>#{index+1}</sup>"])
+              value = sprintf(FOOTNOTE_BACKLINK_FMT, nbsp, "#{name}:#{index}",
+                              "#{backlink_text}<sup>#{index + 1}</sup>")
+              para.children << Element.new(:raw, value)
             end
           end
 
           ol.children << Element.new(:raw, convert(li, 4))
           i += 1
         end
-        (ol.children.empty? ? '' : format_as_indented_block_html('div', {:class => "footnotes"}, convert(ol, 2), 0))
+        if ol.children.empty?
+          ''
+        else
+          format_as_indented_block_html('div', {class: "footnotes", role: "doc-endnotes"}, convert(ol, 2), 0)
+        end
       end
 
     end
